/* Stats - IRC Statistical Services for Hybrid, csircd, and IRCnet 2.10
** users.c - Functions to track network state
**
** Copyright W. Campbell and others.  See README for more details
** Some code Copyright: Jonathan George, Kai Seidler, ircd-hybrid Team,
**                      IRCnet IRCD developers.
**
** $Id: users.c,v 1.105 2002/07/27 01:00:59 wcampbel Exp $
*/

#include <assert.h>
#include "stats.h"
#include "struct.h"
#include "proto.h"
#include "match.h"

/* These are ugly functions, gets right down to the hash tables and
** manipulation of raw data
*/

static struct Server *new_server(char *, char *);
struct Server *serverlist[S_TABLE_SIZE];

struct MyUser *myuhead;

static struct User *new_user(char *);
static void add_user_to_hash_table(char *, struct User *);
static void del_user_from_hash_table(struct User *);
static void add_server_to_hash_table(char *, struct Server *);
static void del_server_from_hash_table(struct Server *);

struct User *userlist[U_TABLE_SIZE];

static void add_user_to_hash_table(char *nick, struct User *u)
{
	u->hash = HASH((unsigned char *) nick, U_TABLE_SIZE);
	u->next = userlist[u->hash];
	userlist[u->hash] = (void *)u;
}

static void del_user_from_hash_table(struct User *u)
{
	struct User *tmp, *prev = NULL;

	for (tmp = userlist[u->hash]; tmp; tmp = tmp->next)
	{
		if (tmp == u)
		{
			if (prev)
				prev->next = tmp->next;
			else
			{
				/* This check shouldn't be necessary...
				** but Linux seems to care...
				*/
				if (tmp->next)
					userlist[u->hash] = tmp->next;
				else
					userlist[u->hash] = NULL;
			}
			tmp->next = NULL;
			return;
		}
		prev = tmp;
	}
}

static void add_server_to_hash_table(char *name, struct Server *s)
{
	s->hash = HASH((unsigned char *) name, S_TABLE_SIZE);
	s->next = serverlist[s->hash];
	serverlist[s->hash] = (void *)s;
        debug("add_server_to_hash_table: %s, %d", name, s->hash);
}

static void del_server_from_hash_table(struct Server *s)
{
	struct Server *tmp, *prev = NULL;

	for (tmp = serverlist[s->hash]; tmp; tmp = tmp->next)
	{
		if (tmp == s)
		{
			if (prev)
				prev->next = tmp->next;
			else
				serverlist[s->hash] = tmp->next;
			tmp->next = NULL;
			return;
		}
		prev = tmp;
	}
}

static struct User *new_user(char *nick)
{
	struct User *u;

	u = smalloc(sizeof(struct User));
	if (!nick)
		nick = "";
	strlcpy(u->nick, nick, MAXNICK + 1);
	add_user_to_hash_table(nick, u);

	return u;
}

#ifdef IRCD_2_10
void AddUser(char *nick, char *user, char *host, struct Server *server)
#else
void AddUser(char *nick, char *user, char *host, char *server)
#endif
{
	struct User *u;

#ifdef IRCD_2_10
	debug("AddUser(): %s (%s@%s) -> %s", nick, user, host,
			server->name);
#else
	debug("AddUser(): %s (%s@%s) -> %s", nick, user, host, server);
#endif
	u = finduser(nick);
	if (u)
	{
		log("trying to add a user that already exists? (%s)", nick);
		return;
	}

	u = new_user(nick);
	strlcpy(u->hostname, host, MAXHOST);
	strlcpy(u->username, user, MAXUSER);
#ifdef IRCD_2_10
	u->server = server;
#else
	u->server = findserver(server);
#endif
	u->t_flood = time(NULL);
	u->tld = AddTLD(u);

	if (!u->server)
		log("Unable to find server for user %s (%s?)", nick, server);
	else
		IncreaseUsers(u->server);
}

void DelUser(char *nick)
{
	struct User *u = finduser(nick);

	debug("DelUser(%s)", nick);

	if (!u)
	{
		log("DelUser(%s) failed!", nick);
		return;
	}

	del_user_from_hash_table(u);

	DecreaseUsers(u->server);

	if (u->tld)
		u->tld->users--;
	if (IsAway(u))
	{
		netinfo.away--;
#if defined(IRCD_HYB7) || defined(IRCD_2_10)
		u->server->away--;
#endif
	}
	if (IsOper(u))
	{
		netinfo.opers--;
		u->server->opers--;
	}
	if (u->myuser)
		u->flags &= ~MFLAG_ISON;

	/* Debug asserts */
	assert(u->tld->users >= 0);
	assert(u->server->opers >= 0);
	assert(netinfo.opers >= 0);
	assert(netinfo.users >= 0);

	free(u);
}

void Change_User(struct User *u, char *newnick)
{
	debug("Change_User(%s, %s)", u->nick, newnick);

	spy("NICK Change by %s to %s", u->nick, newnick);

	del_user_from_hash_table(u);

	u->nick[1] = 0;
	strlcpy(u->nick, newnick, MAXNICK + 1);

	add_user_to_hash_table(u->nick, u);
}

struct User *finduser(char *nick)
{
	struct User *u;

	debug("finduser(%s)", nick);

	u = userlist[HASH((unsigned char *) nick, U_TABLE_SIZE)];
	while (u && (strcasecmp(u->nick, nick) != 0))
		u = u->next;

	debug("finduser(%s) -> %s", nick, (u) ? u->nick : "NOTFOUND");

	return u;
}

#ifdef IRCD_2_10
struct Service *findservice(char *nick, char *server)
{
	struct Service *s = NULL;
	dlink_node *dl;

	debug("findservice(%s, %s)", nick, server);

	for (dl = servicehead.head; dl; dl = dl->next)
	{
		s = (struct Service *) dl->data;
		if (!(strcasecmp(nick, s->nick)) &&
		    !(strcasecmp(server, s->server)))
			return s;
	}

	return NULL;
}

void AddService(char *nick, char *server, char *dist, char *gecos)
{
	struct Service *s = smalloc(sizeof(struct Service));
	dlink_node *dl = smalloc(sizeof(dlink_node));
	struct Server *sv = findserver(server);
	time_t current = time(NULL);

	strlcpy(s->nick, nick, MAXNICK + 1);
	strlcpy(s->server, server, MAXHOST);
	strlcpy(s->dist, dist, MAXHOST);
	strlcpy(s->gecos, gecos, LINE);
	dlinkAdd(s, dl, &servicehead);
	netinfo.services++;
        if (netinfo.services > netinfo.maxservices)
        {
                netinfo.maxservices = netinfo.services;
                netinfo.t_maxservices = current;
        }
	if (netinfo.services > daily.services)
	{
		daily.services = netinfo.services;
		daily.t_services = current;
	}
	if (sv)
	{
		sv->services++;
		if (sv->services > sv->stats->maxservices)
		{
			sv->stats->maxservices = sv->services;
			sv->stats->t_maxservices = current;
		}
	}
}

void DelService(char *nick, char *server)
{
	struct Service *s = NULL;
	struct Server *sv = findserver(server);
	dlink_node *dl;

	for (dl = servicehead.head; dl; dl = dl->next)
	{
		s = (struct Service *) dl->data;
		if (!(strcasecmp(nick, s->nick)) &&
		    !(strcasecmp(server, s->server)))
		{
			dlinkDelete(dl, &servicehead);
			free(s);
			free(dl);
			netinfo.services--;
                        if (sv)
                                sv->services--;
                        return;
		}
	}

	log("Unable to delete service %s@%s", nick, server);
}

void DelServicesByServer(char *server)
{
	struct Service *s = NULL;
	dlink_node *dl = servicehead.head;
	dlink_node *tdl = NULL;
	struct Server *sv = findserver(server);

	while (dl != NULL)
	{
		s = (struct Service *) dl->data;
		if (!(strcasecmp(server, s->server)))
		{
			tdl = dl->next;
			dlinkDelete(dl, &servicehead);
			free(dl);
			free(s);
			netinfo.services--;
                        /* We're getting rid of this server record...
                        ** is this really needed?
                        */
                        if (sv)
                                sv->services--;
			dl = tdl;
		}
		else
		{
			dl = dl->next;
		}
	}
}
			
#endif

void init_user_hash()
{
	int i;
	struct User *u, *prev;

	for (i = 0; i < U_TABLE_SIZE; i++)
	{
		u = userlist[i];
		while (u != NULL)
		{
			prev = u->next;
			debug("init_user_hash() deleting %s", u->nick);
			free(u);

			u = prev;
		}
		userlist[i] = NULL;
	}
	memset((void *)userlist, '\0', sizeof(userlist));
}

void UserMode(char *nick, char *modes)
{
	struct User *u = finduser(nick);
	int add = 0;

	if (!u)
	{
		log("Changing modes for unknown user: %s", nick);
		return;
	}

	while (*modes != '\0')
	{
		switch(*modes)
		{
			case '+': add = 1;	break;
			case '-': add = 0;	break;
			case 'o':
				if (add)
				{
				    if (!IsOper(u))
				    {
					u->flags |= UFLAG_ISOPER;
					IncreaseOpers(u->server);
#ifdef OPER_MODES
					u->operflags |= OPER_GLOBAL;
#endif
					alert("New Oper %s on %s", u->nick,
						u->server->name);
				        debug("%s is now an IRCop.", u->nick);
				    }
				}
				else
				{
				    if (IsOper(u))
				    {
					u->flags &= ~UFLAG_ISOPER;
					netinfo.opers--;
					u->server->opers--;
#ifdef OPER_MODES
					u->operflags = 0;
#endif
					alert("Deoper by %s on %s", u->nick,
						u->server->name);
					debug("%s is no longer an IRCop.",
						u->nick);
					assert(netinfo.opers >= 0);
					assert(u->server->opers >= 0);
				    }
				}
				break;
#ifdef IRCD_2_10
			/* IRCnet propegates AWAY by setting usermode +a */
			case 'a':
				if (add)
				{
				    if (!IsAway(u))
				    {
					u->flags |= UFLAG_ISAWAY;
					netinfo.away++;
					u->server->away++;
					if (u->server->away >
						u->server->stats->maxaway)
					{
					    u->server->stats->maxaway =
						u->server->away;
					    u->server->stats->t_maxaway =
						time(NULL);
					}
					if (netinfo.away > netinfo.maxaway)
					{
					    netinfo.maxaway = netinfo.away;
					    netinfo.t_maxaway = time(NULL);
					}
				    }
				}
				else
				{
				    if (IsAway(u))
				    {
					u->flags &= ~UFLAG_ISAWAY;
					netinfo.away--;
					u->server->away--;
				    }
				    assert(netinfo.away >= 0);
				}
				break;
#endif
#ifdef OPER_MODES
			case 'a':
				if (add)
					u->operflags |= OPER_ADMIN;
				else
					u->operflags &= ~OPER_ADMIN;
				if (add)
					alert("New Server Admin %s on %s",
						u->nick, u->server->name);
				else
					alert("Server Admin Status Removal %s "
						"on %s", u->nick,
						u->server->name);
				break;
			case 'O':
				if (add)
					u->operflags |= OPER_LOCAL;
                                else
                                        u->operflags &= ~OPER_LOCAL;
                                break;
#endif /* OPER_MODES */
			default:
				break;
		}
		modes++;
	}
}

static struct Server *new_server(char *name, char *desc)
{
	struct Server *s;

	s = calloc(1, sizeof(struct Server));
	/* No need to memset here... calloc zeros out all
	** memory it allocates
	*/
	if (!name)
		name = "";
	strlcpy(s->name, name, MAXHOST);
	if (!desc)
		desc = "";
#ifndef IRCD_2_10
	/* Remove the (H) hidden identifier */
	if ( (desc[0] == '(') &&
             (desc[1] == 'H') &&
             (desc[2] == ')') )
		desc += 4;
#endif
	strlcpy(s->desc, desc, LINE);
	add_server_to_hash_table(name, s);

	debug("new_server: %s, %s", name, desc);

	return s;
}

#ifdef IRCD_2_10
void AddServer(char *name, int hops, int token, char *desc, char* uplink)
#else
void AddServer(char *name, int hops, char *desc, char* uplink)
#endif
{
	struct Server *s;
	time_t current = time(NULL);

#ifdef JUPESERV
#if defined(TRACK_ADMIN) || defined(TRACK_VERSION)
	struct Jupe *j = FindJupe(name);
#endif
#endif

	debug("New Server: %s", name);
	strlower(name);

	/* Check to see if a server already exists - Why is a mystery to me 
	** but sometimes on Hyb7 it'll happen
	*/
	s = findserver(name);
	if (s)
	{
		log("Server %s already exists in the server table", name);
	}
	else
	{
		s = new_server(name, desc);
	}
	s->hops = hops;
#ifdef IRCD_2_10
	s->token = token;
#endif
	s->connected_since = current;
	s->last_announce = current;
#ifdef LAG_CHECK
	s->ping = -1;
	/* new_server() callocs, so most of these are already set to 0 */
	s->t_lowest_ping = current;
	s->t_highest_ping = current;
#endif /* LAG_CHECK */
        if (uplink != NULL)
                s->uplink = findserver(uplink);

	debug("findserver() --> %s uplink %s", s->name,
		s->uplink ? s->uplink->name : "NULL");

#ifdef TRACK_VERSION
	/* If it is NOT settings.name */
	if (strcasecmp(settings.name, name))
	{
#ifdef JUPESERV
		if (j == NULL)
#endif
			sts(":%s VERSION %s", AdminServ.nick, name);
	}
	else
	{
		snprintf(s->version, BUFSIZE, "%s %s %s -> %s",
			stats_version, settings.name, VersionFlags(),
			stats_date);
	}
#endif

#ifdef TRACK_ADMIN
        /* If it is NOT settings.name */
        if (strcasecmp(settings.name, name))
        {
#ifdef JUPESERV
		if (j == NULL)
#endif
                	sts(":%s ADMIN %s", AdminServ.nick, name);
		/* admin items do not need to be zero'ed, calloc'ed memory */
        }
        else
        {
                snprintf(s->admin1, BUFSIZE, "%s Network Statistical Services",
                        settings.netname);
		snprintf(s->admin2, BUFSIZE, "%s", settings.admin_name);
		snprintf(s->admin3, BUFSIZE, "%s", settings.admin_email);
        }
#endif

	/* Check to see if it should be hidden */
	if (findhidden(name))
		s->flags |= SFLAG_HIDE;

	/* Check for the (H) in the server desc field, and mark the
	** server as hidden.  We probably should not add it to the 
	** Hidden list though, it'll be a waste
	*/

#ifndef IRCD_2_10
	/* Keep it enabled even if we are not linked to a Hybrid 7
	** server, just in case there's a "hidden hub" on the network
	*/
	if (desc)
		if ( (desc[0] == '(') &&
		     (desc[1] == 'H') &&
                     (desc[2] == ')') )
			s->flags |= SFLAG_HIDE;
#endif
	
	IncreaseServers();
	AddStats(s);
}

/* This is similar to DelServer, but always recurses down the tree!
** This function is used when the connection to the uplink is lost
*/
void FlushServer(char *name)
{
        struct Server *up = NULL;
        struct User* u = NULL;
	char *tstr = NULL;
	struct Server *s = findserver(name);
        int i=0;
	debug("FlushServer called for %s", name);

	if (!s || !name)
	{
		log("FlushServer(): %s failed!", name);
                return;
        }

        u=userlist[i];
        while (TRUE)
	{
                if (i == U_TABLE_SIZE)
		{
                        break;
                }
                if (u)
		{
                        if (!strcmp(name, u->server->name))
			{
                                if (u->next)
				{
                                        /* UGH kludge */
                                        tstr = u->nick;
                                        u = u->next;
                                        DelUser(tstr);
                                        tstr = NULL;
                                        continue;
                                }
				else
				{
                                        DelUser(u->nick);
					i++;
					u = userlist[i];
					continue;
                                }
                        }

                        if (u->next)
			{
                                u = u->next;
                        }
			else
			{
                                i++;
                                u = userlist[i];
                        }
                }
		else
		{
			i++;
			u = userlist[i];
		}
        }
        i = 0;
        up = serverlist[i];
        while (TRUE)
	{
                if (i == S_TABLE_SIZE)
		{
                        break;
                }
                if (up)
		{
                        if (up->uplink)
			{
                                if (!strcmp(name, up->uplink->name))
				{
                                        if (up->next)
					{
                                                tstr = up->name;
                                                up = up->next;
                                                FlushServer(tstr);
                                                tstr = NULL;
                                                continue;
                                        }
					else
					{
                                                FlushServer(up->name);
						i++;
						up = serverlist[i];
						continue;
                                        }
                                }
                        }
                        if (up->next)
			{
                                up = up->next;
                        }
			else
			{
                                i++;
                                up = serverlist[i];
                        }
                }
		else
		{
                        i++;
                        up = serverlist[i];
                }
        }

#ifdef SPLITSERV
	SpDelServer(name);
#ifdef TRACK_SPLITS_ON_DISCONNECT
	if (!IsSPHide() || !((IsSPHide() && IsSHide(s)) ||
		(IsSPHide() && IsSHide(s->uplink))))
		SpNewSplit(name);
#endif
#endif

        del_server_from_hash_table(s);
        s->stats->lastseen = time(NULL);
        DecreaseServers();
        free(s);
}

void DelServer(char *name)
{
	struct Server *s = findserver(name);

#if defined(IRCD_HYB6) || defined(IRCD_HYB7) || defined(IRCD_2_10)
	/* This is to support the QS capab
	** IRCnet 2.10.x does this too
	*/
	struct Server *up = NULL;
	struct User* u = NULL;
	char *tstr = NULL;
	int i=0;
	u=userlist[i];
#endif

        if (!s || !name)
        {
                log("DelServer(): %s failed!", name);
                return;
        }

        if (s)
                s->stats->numsplits++;

#ifdef IRCD_2_10
        DelServicesByServer(name);
#endif

#if defined(IRCD_HYB6) || defined(IRCD_HYB7) || defined(IRCD_2_10)
	while (TRUE)
	{
		if (i == U_TABLE_SIZE)
		{
			break;
		}
		if (u)
		{
			if (!strcmp(name, u->server->name))
			{
				if (u->next)
				{
					/* UGH kludge */
					tstr = u->nick;
					u = u->next;
					DelUser(tstr);
					tstr = NULL;
					continue;
				}
				else
				{
					DelUser(u->nick);
					i++;
					u = userlist[i];
					continue;
				}
			}

			if (u->next)
			{
				u = u->next;
			}
			else
			{
				i++;
				u = userlist[i];
			}
                }
		else
		{
                        i++;
                        u = userlist[i];
                }
	}
	i = 0;
	up = serverlist[i];
	while (TRUE)
	{
		if (i == S_TABLE_SIZE)
		{
			break;
		}
		if (up)
		{
			if (up->uplink)
			{
				if (!strcmp(name, up->uplink->name))
				{
					if (up->next)
					{
						tstr = up->name;
						up = up->next;
						DelServer(tstr);
						tstr = NULL;
						continue;
					}
					else
					{
						DelServer(up->name);
						i++;
						up = serverlist[i];
						continue;
					}
				}
			}
			if (up->next)
			{
				up = up->next;
			}
			else
			{
				i++;
				up = serverlist[i];
			}
		}
		else
		{
			i++;
			up = serverlist[i];
		}
	}
#endif

#ifdef SPLITSERV

#if defined(IRCD_HYB6) || defined(IRCD_HYB7) || defined(IRCD_2_10)
                        splog("Split: %s from %s", name,
                                (s && (s->uplink)) ? s->uplink->name
                                        : "UNKNOWN");
                        if (!is_delayed)
                        {
                                if (IsSPReport())
				{
				  if (!IsSPHide() ||
					!((IsSPHide() && IsSHide(s)) ||
					(IsSPHide() && IsSHide(s->uplink))))
                                    privmsg(sp.join_chan, SplitServ.nick,
					MSG_SPDEFAULT,
                                        "SPLIT: %s from %s", name,
                                        (s && (s->uplink)) ? s->uplink->name :
                                        "UNKNOWN");
				}
                        }
#else /* IRCD_HYB6 */

                        /* Correct this for reasons */
                        splog("Split: %s", name);
                        if (!is_delayed)
                        {
                                if (IsSPReport())
				{
				  if (!IsSPHide() ||
					!((IsSPHide() && IsSHide(s)) ||
					(IsSPHide() && IsSHide(s->uplink))))
                                    privmsg(sp.join_chan, SplitServ.nick,
					MSG_SPDEFAULT,
                                        "SPLIT: %s", name);
				}
                        }
#endif /* IRCD_HYB6 */

        SpDelServer(name);
	if (!IsSPHide() || !((IsSPHide() && IsSHide(s)) ||
		(IsSPHide() && IsSHide(s->uplink))))
        	SpNewSplit(name);
#endif

	del_server_from_hash_table(s);
	s->stats->lastseen = time(NULL);
	DecreaseServers();
	assert(netinfo.servers >= 0);
	free(s);
}

#ifdef IRCD_2_10
struct Server *findservtoken(int token)
{
        struct Server *s;
        int i;

        for(i=0; i<S_TABLE_SIZE; i++)
	{
                s = serverlist[i];
		while (s && (s->token != token))
			s = s->next;
		if (s)
			return s;
        }

	/* We should never get here */
       	debug("findservtoken(%d) -> %s", token, (s) ? s->name : "NOT FOUND");

	/* I'd rather return an explicit NULL here */
        return NULL;
}
#endif

struct Server *findserver(char *name)
{
	struct Server *s;
	int i;

	strlower(name); 
	debug("findserver(%s)", name);
	s = serverlist[HASH((unsigned char *) name, S_TABLE_SIZE)];
	while (s && (strcasecmp(s->name, name) != 0))
		s = s->next;
	if (s)
		return s;

/* The name in the server table is the pattern (with the *) 
** The name is what is passed into this function (usually full 
** server name)
**
** NOTE:  This also is a performance hit, if there are a number of masked
**        servers or there are a lot of users on that masked server/set of
**        servers hidden by the mask
*/
	for(i=0; i<S_TABLE_SIZE; i++)
	{
		s = serverlist[i];
        	while (s && (match(s->name, name) != 0))
                	s = s->next;
		if (s)
			return s;
	}

	/* We should never get here unless the server doesn't exist */
	debug("findserver(%s) -> %s", name, (s) ? s->name : "NOT FOUND");

/* I'd rather return an explicit NULL here */
	return NULL;
}

void init_server_hash()
{
	int i;
	struct Server *s, *b;

	for (i = 0; i < S_TABLE_SIZE; i++)
	{
		s = serverlist[i];
		while (s)
		{
			b = s->next;
			debug("init_server_hash() deleting %s", s->name);
			free(s);
			s = b;
		}
		serverlist[i] = NULL;
	}
	memset((void *)serverlist, '\0', sizeof(serverlist));
#ifdef IRCD_2_10
	AddServer(settings.name, 0, 0, settings.desc, NULL);
#else
	AddServer(settings.name, 0, settings.desc, NULL);
#endif
}

void LoadMyUsers()
{
	FILE *fp;
	int parc;
	int i;
	char *parv[MAXPARA];
	char buf[BUFSIZE];
	struct MyUser *myu;

        /* Clear the parv */
        for (i = 0; i < MAXPARA; i++)
                parv[i] = NULL;

	if ((fp = fopen(USERFILE, "r")) == NULL)
	{
		fprintf(stderr, "Please type this: echo \"<nickname> "
			"<password> 0 >> %s\"\n", USERFILE);
		fprintf(stderr, "Replace nickname with your nickname, and "
			"password with your password, encrypting if "
			"necessary\n");
		fprintf(stderr, "This is used for the LOGIN feature of "
			"Stats.\n");
		log("Unable to open userlist.");
		exit(0);
	}

	while (fgets(buf, BUFSIZE, fp))
	{
		parc = msgtoken(buf, parv);

		if ((!parv[0]) || (!parv[1]))
		{
			fprintf(stderr, "Invalid entry in user file: %s\n",
				USERFILE);
			fprintf(stderr, "Format: nick pass 0\n");
			fprintf(stderr, "Or: nick pass <time last seen>\n");
			fprintf(stderr, "Exiting...\n");
			exit(0);
		}

                myu = smalloc(sizeof(struct MyUser));
                strlcpy(myu->username, parv[0], MAXNICK + 1);
                strlcpy(myu->password, parv[1], LINE);

		if (!parv[2])
		{
			log("Invalid time value in user file %s, defaulting "
				"to 0", USERFILE);
			myu->lastseen = 0;
		}
		else
		{
			myu->lastseen = atol(parv[2]);
		}

		if (!myuhead)
		{
			myuhead = myu;
			myuhead->next = NULL;
		}
		else
		{
			myu->next = myuhead;
			myuhead = myu;
		}
	}
	fclose(fp);
}

void SaveMyUsers()
{
	struct MyUser *myu;
	FILE *fp;

	if (!myuhead)
		return;

	if ((fp = fopen(USERFILE, "w")) == NULL)
	{
		log("Unable to open %s for writing.", USERFILE);
		return;
	}
	for (myu = myuhead; myu; myu = myu->next)
	{
		fprintf(fp, "%s %s %ld\n", myu->username, myu->password,
			myu->lastseen);
	}
	fclose(fp);
}

void DeleteMyUser(char *username)
{
	struct MyUser *myu, *prev = NULL;

	if (!myuhead)
		return;

	myu = myuhead;

	while (myu)
	{
		if (!strcasecmp(myu->username, username))
			break;
		prev = myu;
		myu = myu->next;
	}
	if (!myu)
		return;

	if (!prev)
	{
		myuhead = myu->next;
	}
	else if (myu->next == NULL)
	{
		prev->next = NULL;
	}
	else
	{
		prev->next = myu->next;
	}
	free(myu);
}

struct MyUser *findmyuser(char *nick)
{
	struct MyUser *myu;

	for (myu = myuhead; myu; myu = myu->next)
	{
		if (!strcasecmp(nick, myu->username))
			return myu;
	}
	return NULL;
}
