/* Stats - IRC Statistical Services for Hybrid, csircd, and IRCnet 2.10
** util.c - Misc. utility functions
**
** Copyright W. Campbell and others.  See README for more details
** Some code Copyright: Jonathan George, Kai Seidler, ircd-hybrid Team,
**                      IRCnet IRCD developers.
**
** $Id: util.c,v 1.71 2002/07/28 01:34:17 wcampbel Exp $
*/

#include "stats.h"
#include "struct.h"
#include "proto.h"
#include "setup.h"
#include <signal.h>

/* print a period of seconds into human readable format */
char *timediff(time_t u)
{
        static char buf[TIMEBUF];

        if (u > 86400)
                snprintf(buf, TIMEBUF, "%ld day%s, %02ld:%02ld",
                        u/86400, (u/86400 == 1) ? "" : "s",
                        (u/3600) % 24, (u/60) % 60);
        else if (u > 3600)
                snprintf(buf, TIMEBUF, "%ld hour%s, %ld minute%s",
                        u/3600, u/3600==1 ? "" : "s", (u/60) % 60,
                        (u/60)%60 == 1 ? "" : "s");
        else
                snprintf(buf, TIMEBUF,
                        "%ld minute%s, %ld second%s",
                        u/60, u/60 == 1 ? "" : "s",
                        u%60, u%60 == 1 ? "" : "s");

        return buf;
}

/* Generic logging routine */
void log(char *fmt, ...)
{
	va_list ap;
	FILE *logfile;
	char buf[BUFSIZE], fmtime[LINE];
	time_t tmp = time(NULL);

	va_start(ap, fmt);
	vsnprintf(buf, BUFSIZE, fmt, ap);

	strftime(fmtime, LINE, TIMEFMT, localtime(&tmp));

	if ((logfile = fopen(LOGFILE, "a")) == NULL)
		return;

	if (logfile)
		fprintf(logfile, "(%s) %s\n", fmtime, buf);

	if (IsDebug())
	{
		fprintf(stderr, "(%s) %s\n", fmtime, buf);
	}

	fclose(logfile);
	va_end(ap);
}

#ifdef SPLITSERV
/* SplitServ logging routine */
void splog(char *fmt, ...)
{
        va_list ap;
        FILE *logfile;
        char buf[BUFSIZE], fmtime[LINE];
        time_t tmp = time(NULL);

        va_start(ap, fmt);
        vsnprintf(buf, BUFSIZE, fmt, ap);

        strftime(fmtime, LINE, TIMEFMT, localtime(&tmp));

        if ((logfile = fopen(SPLITLOG, "a")) == NULL)
		return;

        if (logfile)
                fprintf(logfile, "(%s) %s\n", fmtime, buf);

        if (IsDebug())
	{
                fprintf(stderr, "(%s) %s\n", fmtime, buf);
        }

        fclose(logfile);
        va_end(ap);
}
#endif

/* Rotate the logfile, called from chk() */
void ResetLogs()
{
	char tmp[25];
	time_t t = time(NULL);

	/* Fix the date stamp so it's the logs of THAT day, not the previous
	** day with today's date stamp on it
	*/
	t = t - 86400;
	/* It's MUCH easier to manipulate this than all the vars inside the
	** struct
	*/

	strftime(tmp, 25, BACKUPLOGFMT, localtime(&t));
	rename(LOGFILE, tmp);
	log("Started fresh logfile.");

#ifdef SPLITSERV
	strftime(tmp, 25, BACKUPSPLOGFMT, localtime(&t));
	rename(SPLITLOG, tmp);
	splog("Started fresh logfile.");
#endif
}

/* Our own ctime function */
/* Modified for l10n - 12 fv 2002 -Hwy */
char *sctime(time_t stuff)
{
	static char buf[32];
	struct tm *ltm = localtime(&stuff);

	/* %c is the POSIX localized replacement for ctime() */
	strftime(buf, 32, "%c", ltm);
	return buf;
}

/* custom ftime format */
char *sftime(time_t stuff)
{
	static char fmtime[LINE];
	struct tm *ltm = localtime(&stuff);

	if (stuff != 0)
	{
		strftime(fmtime, LINE, FTMFMT, ltm);
	}
	else
	{
		snprintf(fmtime, LINE, "No Time/Record");
	}

	return fmtime;
}

char *irc_time(time_t ts)
{
	static char fmtime[LINE];

	struct tm *ltm = localtime(&ts);

	/* Since we already depend on POSIX local support, %z in the time
	** format string is acceptable...GMToffset was a nice function
	** though :)
	*/
	strftime(fmtime, LINE, IRCTFMT, ltm);

	return fmtime;
}

#if defined(CRYPT) && defined(MD5)
/* Create a random MD5 based salt for password
** encryption
*/
char* make_salt()
{
	static char saltChars[] = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789./";
	static char salt[13];
	int i;

	salt[0] = '$';
	salt[1] = '1';
	salt[2] = '$';
	for (i=3; i<=10; i++)
		salt[i] = saltChars[random() % 64];
	salt[11] = '$';
	salt[12] = '\0';

	return salt;
}
#endif

#if defined(CRYPT) && !defined(MD5)
/* create a random salt for password encryption 
** This only creates DES based salts...
**
** NOTE:  This code is partially derived from mkpasswd.c
**        by Nelson Minar (minar@reed.edu)
*/
char* make_salt()
{
	static char saltChars[] = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789./";
	static char salt[3];

	salt[0] = saltChars[random() % 64];
	salt[1] = saltChars[random() % 64];
	salt[2] = 0;

	return salt;
}
#endif /* CRYPT */

/* tokenize splits apart a message (in our case with the origin and command
** picked off already...
*/
int tokenize(char *message, char **parv)
{
        char *pos = NULL;
        char *next;
        int count = 0;

        if (!message)
        {
                /* Something is seriously wrong...bail */
                return -1;
        }

        /* First we find out if there's a : in the message, save that string
        ** somewhere so we can set it to the last param in parv
	** Also make sure there's a space before it...if not, then we're
	** screwed
        */
	pos = message;
	while (1)
	{
        	if ((pos = strchr(pos, ':')))
        	{
			pos--;
			if (*pos != ' ')
			{
				/* There SHOULDN'T be a problem with this... */
				pos += 2;
				continue;
			}
                	*pos = '\0';
                	/* This is BAD - get rid of the space before the : */
                	pos++;
                	*pos = '\0';
                	/* VERY BAD...over now though */
                	pos++;
			break;
        	}
		else
		{
			break;
		}
	}

        /* Now we take the beginning of message and find all the spaces...
        ** set them to \0 and use 'next' to go through the string
        */

	next = message;
	parv[0] = message;
	count = 1;

	while (*next)
	{
		if (count == MAXPARA - 1)
		{
			/* We've reached one less than our max limit
			** to handle the parameter we already ripped off
			*/
			log("Reached the MAXPARA limit!");
			return count;
		}
		if (*next == ' ')
		{
			*next = '\0';
			next++;
                        /* Eat any additional spaces */
                        while (*next == ' ')
                                next++;
                        parv[count] = next;
                        count++;
                }
                else
                {
                        next++;
                }
        }

        if (pos)
        {
                parv[count] = pos;
                count++;
        }

        return count;
}

/* A generic tokenizer, tokenizes based on a specified character */
int generictoken(char delim, char *message, char **parv)
{
        char *next;
        int count;

        if (!message)
        {
                /* Something is seriously wrong...bail */
                return -1;
        }

        /* Now we take the beginning of message and find all the spaces...
        ** set them to \0 and use 'next' to go through the string
        */
	next = message;
	parv[0] = message;
	count = 1;

        while (*next)
        {
		/* This is fine here, since we don't have a :delimited
		** parameter like tokenize
		*/
        	if (count == MAXPARA)
		{
		    /* We've reached our limit */
		    log("Reached the MAXPARA limit!");
		    return count;
		}
		if (*next == delim)
		{
			*next = '\0';
			next++;
			/* Eat any additional delimiters */
			while (*next == delim)
				next++;
			parv[count] = next;
			count++;
		}
		else
		{
			next++;
		}
        }

	return count;
}

/* our own malloc with logging and sanity checking */
void *smalloc(size_t size)
{
        void *buf;

        if (!size)
	{
                log("smalloc(): illegal attempt to allocate 0 bytes!");
                size = 1;
        }
        buf = malloc(size);
        if (!buf)
	{
                log("smalloc(): out of memory.");
                exit(0);
        }
	memset((void *) buf, 0, size);
        return buf;
}

/* Generate a hash value into the tables */
unsigned long HASH(const unsigned char *name, int size_of_table)
{
        unsigned long h = 0, g;

        while (*name)
	{
                h = (h << 4) + *name++;
                        if ((g = (h & 0xF0000000)))
                                h ^= g >> 24;
                        h &= ~g;
        }
        return h % size_of_table;
}

/* Lowercase a string */
char *strlower(char *s)
{
        char *t = s;
        while ((*t++ = tolower(*t)))
                ;
        return s;
}

/* Get the version flags string */
char *VersionFlags()
{
	static char ver[LINE];
	char *verptr = ver;

#ifdef TRACK_ADMIN
	*verptr++ = 'A';
#endif
#ifdef SUPPORT_MSG_UH
	*verptr++ = 'U';
#endif
#ifdef SUPPORT_MSG_OPERS
	*verptr++ = 'O';
#endif
#ifdef IGNORE_ILLEGAL_TLD
	*verptr++ = 'I';
#endif
#ifdef TRACK_VERSION
	*verptr++ = 'V';
#endif
#ifdef LAG_CHECK
	*verptr++ = 'L';
#endif
#ifdef JUPESERV
	*verptr++ = 'J';
#endif
#ifdef ENABLE_QUOTE
	*verptr++ = 'Q';
#endif
#ifdef DELAY_PRIVMSG
	*verptr++ = 'D';
#endif
#ifdef LARGE_NET
	*verptr++ = 'N';
#endif
#ifdef OPER_MODES
	*verptr++ = 'M';
#endif
#ifdef DIE_ON_MASS_COLLIDE
	*verptr++ = 'C';
#endif
#ifdef DISALLOW_NICK_PRIVMSG
	*verptr++ = 'u';
#endif
#ifdef SPAMSERV
	*verptr++ = 'S';
#endif
#ifdef SPLITSERV
	*verptr++ = 'P';
#endif
#ifdef STATSERV
	*verptr++ = 'T';
#endif
#ifdef ALERTMODE
	*verptr++ = 'a';
#endif
#ifdef SPYMODE
	*verptr++ = 's';
#endif

	if (IsDebug())
		*verptr++ = 'd';
	if (IsNoFork())
		*verptr++ = 'n';

	*verptr++ = ' ';
	*verptr = '\0';

#ifdef IRCD_HYBRID
	strlcat(ver, "HYBRID", LINE);
#endif
#ifdef IRCD_HYB6
	strlcat(ver, "HYB6", LINE);
#endif
#ifdef IRCD_HYB7
	strlcat(ver, "HYB7", LINE);
#endif
#ifdef IRCD_2_10
	strlcat(ver, "2_10", LINE);
#endif

	return ver;
}

#ifdef OPER_MODES
char opermode(struct User *u)
{
	if (u == NULL)
		return '-';
	if (u->operflags & OPER_ADMIN)
		return 'A';
	if (u->operflags & OPER_GLOBAL)
		return 'O';
	if (u->operflags & OPER_LOCAL)
		return 'o';
	/* Shouldn't get here */
	return 'x';
}
#endif /* OPER_MODES */

void SigHup(int sig)
{
	need_rehash = 1;
	signal(SIGHUP, SigHup);
}

void SigUsr1(int sig)
{
	/* This should be sufficient to tell it to reconnect */
	close(servsock);
	signal(SIGUSR1, SigUsr1);
}

void SigTerm(int sig)
{
        wallops(settings.name, "Received SIGTERM: Exiting");
	alert("Received SIGTERM: Exiting");
#ifdef STATSERV
        sts(":%s QUIT :Received SIGTERM: Exiting", StatServ.nick);
#endif
	sts(":%s QUIT :Received SIGTERM: Exiting", AdminServ.nick);
#ifdef SPLITSERV
	sts(":%s QUIT :Received SIGTERM: Exiting", SplitServ.nick);
#endif
#ifdef SPAMSERV
	sts(":%s QUIT :Received SIGTERM: Exiting", s_Spam);
#endif
        sleep(1);
        close(servsock);
        SaveStats();
        log("Received SIGTERM: Exiting");
	/* No need to reinstall the signal handler */
        exit(0);
}

int is_channel(char *chan)
{
        if (chan)
	{
                if ( (chan[0] == '#') || (chan[0] == '+') )
                        return TRUE;
                return FALSE;
        }

        /* Shouldn't get here */
        return FALSE;
}

void alert(char *fmt, ...)
{
#ifdef ALERTMODE
	va_list ap;
	char buf[BUFSIZE];

	if (is_delayed)
		return;
	va_start(ap, fmt);
	vsnprintf(buf, sizeof(buf), fmt, ap);
	/* snprintf will always NULL terminate */
	privmsg(ALERTCHAN, AdminServ.nick, MSG_PRIVMSG, "%s", buf);
	va_end(ap);
#endif
}

void spy(char *fmt, ...)
{
#ifdef SPYMODE
        va_list ap;
        char buf[BUFSIZE];

	if (is_delayed)
		return;
        va_start(ap, fmt);
        vsnprintf(buf, sizeof(buf), fmt, ap);
        privmsg(SPYCHAN, AdminServ.nick, MSG_PRIVMSG, "%s", buf);
        va_end(ap);
#endif
}

void debug(char *fmt, ...)
{
	va_list ap;
	char buf[BUFSIZE];

	if (IsDebug())
	{
		va_start(ap, fmt);
		vsnprintf(buf, sizeof(buf), fmt, ap);
		log("DEBUG: %s", buf);
		va_end(ap);
	}
}

/*
 * strlcat and strlcpy were ripped from openssh 2.5.1p2
 * They had the following Copyright info:
 *
 *
 * Copyright (c) 1998 Todd C. Miller <Todd.Miller@courtesan.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED `AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */



#ifndef HAVE_STRLCAT
size_t strlcat(char *dst, const char *src, size_t siz)
{
        char *d = dst;
        const char *s = src;
        size_t n = siz, dlen;

        while (*d != '\0' && n-- != 0)
                d++;
        dlen = d - dst;
        n = siz - dlen;

        if (n == 0)
                return(dlen + strlen(s));
        while (*s != '\0') {
                if (n != 1) {
                        *d++ = *s;
                        n--;
                }
                s++;
        }
        *d = '\0';
        return(dlen + (s - src));       /* count does not include NUL */
}
#endif

#ifndef HAVE_STRLCPY
size_t strlcpy(char *dst, const char *src, size_t siz)
{
        char *d = dst;
        const char *s = src;
        size_t n = siz;
        /* Copy as many bytes as will fit */
        if (n != 0 && --n != 0) {
                do {
                        if ((*d++ = *s++) == 0)
                                break;
                } while (--n != 0);
        }
        /* Not enough room in dst, add NUL and traverse rest of src */
        if (n == 0) {
                if (siz != 0)
                        *d = '\0';              /* NUL-terminate dst */
                while (*s++)
                        ;
        }

        return(s - src - 1);    /* count does not include NUL */
}
#endif

void InitLists(void)
{
	hiddenhead.head = NULL;
	hiddenhead.tail = NULL;
#ifdef IRCD_2_10
        servicehead.head = NULL;
	servicehead.tail = NULL;
#endif
}

void ExpandParv(char *dest, int max, int initial, int parc, char *parv[])
{
	int i;

	if (parv[initial])
		strlcpy(dest, parv[initial], max);
	for (i = initial + 1; i < parc; i++)
	{
		strlcat(dest, " ", max);
		strlcat(dest, parv[i], max);
	}
}
